/* 
 * Copyright (c) 1996-1997 by Sun Microsystems, Inc.
 */

#ident "@(#)cbqif.c  1.31     97/04/28 SMI"

/* 
 * Copyright (c) Sun Microsystems, Inc.  1996. All rights reserved. 
 *
 * License is granted to copy, to use, and to make and to use derivative 
 * works for research and evaluation purposes, provided that Sun Microsystems is
 * acknowledged in all documentation pertaining to any such copy or derivative
 * work. Sun Microsystems grants no other licenses expressed or implied. The
 * Sun Microsystems  trade name should not be used in any advertising without
 * its written permission.
 *
 * SUN MICROSYSTEMS MERCHANTABILITY OF THIS SOFTWARE OR THE SUITABILITY OF
 * THIS SOFTWARE FOR ANY PARTICULAR PURPOSE.  The software is provided "as is"
 * without express or implied warranty of any kind.
 *
 * These notices must be retained in any copies of any part of this software.
 */

#include <stdio.h>
#include <stddef.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <cbqd.h>

/*
 * Define local Macros
 */
#define CTL_PBANDWIDTH	2
#define NS_PER_MS	(1000000.0)
#define NS_PER_SEC	(NS_PER_MS*1000.0)
#define RM_FILTER_GAIN	5
#define RM_MAXPRIO	8

#define	CBQ_DEVICE	"/dev/cbq"

#define ADMIT_FLOW_OK		0
#define ADMIT_FLOW_NOSVC	1
#define ADMIT_FLOW_NOBW		2
#define ADMIT_FLOW_SYSERR	3

/* 
 * External declarations 
 */
char           *fmt_spec();
double          pow(double, double);
double          fabs(double);

/*
 * Static globals
 */

static cbq_ifinfo_t	*cbq_iflist = NULL;
static int		cbqfd = -1;
static int		cbqdebug = 0;

/*
 * int
 * cbq_init(int debug) - Open the CBQ device and set the debug level
 *
 *	Returns:	0, if successful.
 *			1, on error.
 */

int
cbq_init(int debug)
{
	if ((cbqfd = open(CBQ_DEVICE, O_RDWR)) < 0) {
		perror("CBQ open");
		return (1);
	}

	cbqdebug = debug;
	return (0);
}

/*
 * int
 * cbq_add_if(char *ifname, long bandwidth, int is_shared, int
 *	is_stats_enabled, int is_sched, int is_wrr, int efficient)
 *
 *	Add a new interface to the list of interfaces that the CBQ
 *	driver is managing.
 *
 *	Returns:	0, if successful.
 *			1, on error.
 */

int
cbq_add_if(char *ifname, long bandwidth, int is_shared,
	int is_stats_enabled, int is_sched, int is_wrr, int efficient)
{
	cbq_ifinfo_t   *ifinfo;
	
	if ((ifinfo = (cbq_ifinfo_t *)malloc(sizeof(cbq_ifinfo_t))) == NULL)
		goto err_ret;

	ifinfo->cllist = NULL;
	ifinfo->is_wrr = is_wrr;
	ifinfo->ifname = (char *)strdup(ifname);
	ifinfo->bandwidth = bandwidth;
	ifinfo->nsPerByte = 
		((1.0 / (double)ifinfo->bandwidth) * NS_PER_SEC * 8);
	ifinfo->is_shared = is_shared;
	ifinfo->use_cbq = is_sched && (cbqfd >= 0);
	ifinfo->is_stats_enabled = is_stats_enabled;
	ifinfo->is_efficient = efficient;

	/* Link the interface info structure */
	ifinfo->next = cbq_iflist;
	cbq_iflist = ifinfo;
	if (cbq_disable(ifname) || cbq_clear(ifname)) {
		cbq_iflist = cbq_iflist->next;
		goto err_ret;
	}
	return 0;

err_ret:
	if (ifinfo) {
		if (ifinfo->ifname)
			free(ifinfo->ifname);
		free(ifinfo);
	}
	return 1;
}

/*
 * int
 * cbq_enableall(void) - Enable CBQ on interfaces that CBQ driver is
 *	managing.
 *
 *	Returns:	0, on success.
 *			1, on failture.
 */

int
cbq_enableall(void)
{
	cbq_ifinfo_t   *ifinfo;
	for (ifinfo = cbq_iflist; ifinfo; ifinfo = ifinfo->next) {
		if (cbq_enable(ifinfo->ifname))
			return 1;
	}
	return 0;
}

/*
 * int
 * cbq_enable(char *ifname) - Enable CBQ on a particular interface.
 *
 *	Returns:	0, if successful.
 *			1, on error.
 */

int
cbq_enable(char *ifname)
{
	struct cbq_interface cbqif;
	cbq_ifinfo_t   *ifinfo;


	for (ifinfo = cbq_iflist; ifinfo; ifinfo = ifinfo->next) {
		if (!strcmp(ifinfo->ifname, ifname))
			break;
	}
	if (!ifinfo)
		return 1;

	/*
	 * Enable the interface
	 */
	memset((void *) &cbqif, 0, sizeof(cbqif));
	strncpy(cbqif.cbq_ifacename, ifname, IFNAMSIZ);
	cbqif.cbq_ifacelen = strlen(ifname);
	if (ifinfo->use_cbq && ioctl(cbqfd, CBQ_ENABLE, &cbqif) < 0) {
		perror("ioctl CBQ_ENABLE");
		ifinfo->use_cbq = 0;
	}

	/*
	 * Enable Statistics on interface, if set.
	 */
	if (ifinfo->is_stats_enabled) {
		if (ifinfo->use_cbq && ioctl(cbqfd,
				CBQ_ENABLE_STATS, &cbqif) < 0) {
			perror("ioctl CBQ_ENABLE_STATS");
			ifinfo->use_cbq = 0;
			return 1;
		}
	}
	return 0;
}

int
cbq_disableall(void)
{
	cbq_ifinfo_t   *ifinfo;
	for (ifinfo = cbq_iflist; ifinfo; ifinfo = ifinfo->next) {
		if (cbq_disable(ifinfo->ifname))
			return 1;
	}
	return 0;
}

int
cbq_disable(char *ifname)
{
	struct cbq_interface cbqif;
	cbq_ifinfo_t   *ifinfo;


	for (ifinfo = cbq_iflist; ifinfo; ifinfo = ifinfo->next) {
		if (!strcmp(ifinfo->ifname, ifname))
			break;
	}
	if (!ifinfo)
		return 1;

	/* Disable the interface */
	memset((void *) &cbqif, 0, sizeof(cbqif));
	strncpy(cbqif.cbq_ifacename, ifname, IFNAMSIZ);
	cbqif.cbq_ifacelen = strlen(ifname);
	if (ifinfo->use_cbq && ioctl(cbqfd, CBQ_DISABLE, &cbqif) < 0) {
		perror("ioctl CBQ_DISABLE");
		ifinfo->use_cbq = 0;
		return 1;
	}
	return 0;
}

static u_long
cbq_create_class(cbq_ifinfo_t * ifinfo, u_long parent_class,
	u_long borrow_class, u_int pri, u_int is_default, u_int is_ctl,
	u_int bandwidth, u_int maxdelay, u_int maxburst,
	u_int minburst, u_int av_pkt_size, u_int max_pkt_sz)
{
	struct cbq_add_class cl_add;

	int		red = 0;
	double          maxq, maxidle_s, maxidle, minidle, 
			offtime, nsPerByte, ptime, classptime;
	double          g = (1.0 - pow(2, -RM_FILTER_GAIN));
	double          f = ((double) bandwidth / (double) ifinfo->bandwidth);
#ifdef notdef
	double		gton = pow(g, 8 * (8 - pri)); 
#endif
	double		gton = pow(g, maxburst);
	double		gtom = pow(g, minburst);
	double		maxrate;

	/*
	 * Compute other class parameters
	 */
	nsPerByte = ifinfo->nsPerByte / f;
	ptime = (double)av_pkt_size * (double)ifinfo->nsPerByte;
	maxrate = f * ((double)ifinfo->bandwidth / 8.0); 
	classptime = ptime * (1.0 - f) / f;
	maxidle = ((1 / f - 1) * ((1 - gton) / gton));
	maxidle_s = (1 - g);
	if (maxidle > maxidle_s)
		maxidle = ptime * maxidle;
	else
		maxidle = ptime * maxidle_s;
	offtime = classptime * (1.0 + (1.0/(1.0 - g)));
	offtime = offtime * (1 - gtom) / gtom;
	offtime = (offtime * 8.0) / nsPerByte;
	minidle = -((double)max_pkt_sz * (double)ifinfo->nsPerByte) / f;
#ifndef USE_HRTIME
	maxidle = ((maxidle * 8.0) / nsPerByte) * pow(2, RM_FILTER_GAIN);
	maxidle = maxidle / 1000.0;
	minidle = minidle / 1000.0;
	offtime = offtime / 1000.0;
#else
	maxidle = ((maxidle * 8.0)/ nsPerByte);
#endif
	maxq = ((double) maxdelay * NS_PER_MS) / (nsPerByte * av_pkt_size); 
	if (maxq < 10)
		maxq = 10;

	/* create class */
	memset((void *) &cl_add, 0, sizeof(cl_add));

	strncpy(cl_add.cbq_iface.cbq_ifacename, ifinfo->ifname, IFNAMSIZ);
	cl_add.cbq_iface.cbq_ifacelen = strlen(ifinfo->ifname);
	cl_add.cbq_class.parent_class_handle = parent_class;
	cl_add.cbq_class.borrow_class_handle = borrow_class;
	cl_add.cbq_class.priority = pri;
	cl_add.cbq_class.nano_sec_per_byte = (u_int)fabs(nsPerByte);
	cl_add.cbq_class.maxq = (u_int) fabs(maxq);
	cl_add.cbq_class.maxidle = (u_int) fabs(maxidle);
	cl_add.cbq_class.minidle = (int)minidle;
	cl_add.cbq_class.offtime = (u_int) fabs(offtime);
	cl_add.cbq_class.defaultclass = is_default;
	cl_add.cbq_class.ctlclass = is_ctl;
	cl_add.cbq_class.red = red;
	cl_add.cbq_class.wrr = ifinfo->is_wrr;
	
	if (ifinfo->use_cbq && ioctl(cbqfd, CBQ_ADD_CLASS, &cl_add) < 0) {
		perror("ioctl CBQ_ADD_CLASS");
		ifinfo->use_cbq = 0; 
		return 0;
	}

	if (cbqdebug > 5) {
		printf("cbq_create_class.%x:\n", cl_add.cbq_class_handle);
		printf("\tavg packet size is %d\n", av_pkt_size);
		printf("\tbandwidth is %f\n", (double)bandwidth);
		printf("\tmaxrate is %f\n", maxrate);
		printf("\tmax, min, g, gton, gtom = %d, %d, %f, %f, %f\n",
			maxburst, minburst, g, gton, gtom);
		printf("\tclassptime, ptime = %f, %f\n", classptime, ptime);
		printf("\tparent class: 0x%x\n", parent_class);
		printf("\tborrow class: 0x%x\n", borrow_class);
		printf("\tpriority: %d\n", pri);
		printf("\tbandwidth (in nano sec's per byte): %d\n",
			(u_int)fabs(nsPerByte));
		printf("\tmax queue size: %d\n", (u_int) fabs(maxq));
		printf("\tmaxidle: %d\n", (u_int) fabs(maxidle));
		printf("\tminidle: %d\n", (int)minidle);
		printf("\tofftime: %d\n", (u_int)fabs(offtime));
		printf("\tdefault: %d\n", is_default);
		printf("\tctl: %d\n", is_ctl);
		printf("\tred: %d\n", red);
		printf("\twrr: %d\n", ifinfo->is_wrr);
		printf("\tefficient: %d\n", ifinfo->is_efficient);
	}

	return (cl_add.cbq_class_handle);
}

int
cbq_add_static_class(char *ifname, char *clname, char *parent_class,
	char *borrow_class, u_int pri, u_int is_default, u_int pbandwidth,
	u_int maxdelay, u_int maxburst, u_int minburst, u_int av_pkt_size,
	u_int max_pkt_size, cbq_tos_t admission_type,
	cbq_ol_t overlimit_action)
{
	cbq_ifinfo_t   *ifinfo;
	cbq_classinfo_t *parent_cl = NULL, *borrow_cl = NULL, *new_cl;
	u_long          cl_handle;
	long            bandwidth;
	u_int		is_ctl = 0;

	if (!strcmp("ctl_class", clname))
		is_ctl = 1;

	/* get interface info */
	for (ifinfo = cbq_iflist; ifinfo; ifinfo = ifinfo->next) {
		if (!strcmp(ifinfo->ifname, ifname))
			break;
	}
	if (!ifinfo)
		return 1;

	/* Get parent info */
	if (parent_class) {
		for (parent_cl = ifinfo->cllist; parent_cl;
		     parent_cl = parent_cl->next) {
			if (!strcmp(parent_cl->clname, parent_class))
				break;
		}
		if (!parent_cl)
			return (1);
	}

	/* Get borrow info */
	if (borrow_class) {
		for (borrow_cl = ifinfo->cllist; borrow_cl;
		     borrow_cl = borrow_cl->next) {
			if (!strcmp(borrow_cl->clname, borrow_class))
				break;
		}
		if (!borrow_cl)
			return (1);
	}

	/* Create class */
	if (is_default) {
		bandwidth = ((pbandwidth - CTL_PBANDWIDTH) *
			ifinfo->bandwidth) / 100;
	} else
		bandwidth = (pbandwidth * ifinfo->bandwidth) / 100;
	cl_handle = cbq_create_class(ifinfo,
		(parent_cl ? parent_cl->handle : NULL_CLASS_HANDLE),
		(borrow_cl ? borrow_cl->handle : NULL_CLASS_HANDLE),
	      	pri, is_default, is_ctl, bandwidth, maxdelay,
		maxburst, minburst, av_pkt_size, max_pkt_size);
	if (cl_handle == NULL_CLASS_HANDLE)
		goto err_ret;


	/* instantiate class descriptor and init */
	if ((new_cl = (cbq_classinfo_t *)
			malloc(sizeof(cbq_classinfo_t))) == NULL)
		return (1);
	new_cl->clname = (char *)strdup(clname);
	new_cl->handle = cl_handle;
	new_cl->parent_cl = parent_cl;
	new_cl->borrow_cl = borrow_cl;
	new_cl->ifinfo = ifinfo;

	/* Set pointers if a guranteed or predictive non-leaf class */
	switch (admission_type) {
	case CBQ_QOS_CNTR_DELAY:
	case CBQ_QOS_CNTR_LOAD:
		if (ifinfo->cd_class)
			goto err_ret;
		ifinfo->cd_class = new_cl;
		new_cl->acinfo.type = admission_type;
		new_cl->acinfo.bandwidth =
			(ifinfo->bandwidth * pbandwidth) / 100;
		new_cl->acinfo.allocated = 0;
		new_cl->acinfo.maxdelay = maxdelay;
		new_cl->acinfo.maxburst = maxburst;
		new_cl->acinfo.av_pkt_size = av_pkt_size;
		ifinfo->use_cbq = (cbqfd >= 0 );
		break;
	case CBQ_QOS_NONE:
		break;
	default:
		goto err_ret;
	}

	/* link handle in list */
	new_cl->handle = cl_handle;
	new_cl->next = ifinfo->cllist;
	ifinfo->cllist = new_cl;

	/*
	 * If were adding the default traffic class, also add the
	 * ctl traffic class.  This is for RSVP, IGMP, and RIP packets.
	 */
	if (is_default) {
		if (cbq_add_static_class(ifname, "ctl_class", parent_class,
				borrow_class, 6, 0, CTL_PBANDWIDTH,
				maxdelay, maxburst, minburst, av_pkt_size,
				max_pkt_size, admission_type,
				overlimit_action)) {
			goto err_ret;
		}
	}

	return 0;

err_ret:
	if (new_cl) {
		if (new_cl->clname)
			free(new_cl->clname);
		free(new_cl);
	}
	return 1;

}

int
cbq_add_static_filter(char *ifname, char *clname, cbq_filt_t *filt)
{
	int             i, j;
	cbq_ifinfo_t   *ifinfo;
	cbq_classinfo_t *class = NULL;
	struct cbq_add_filter fltr_add;

	/* get interface info */
	for (ifinfo = cbq_iflist; ifinfo; ifinfo = ifinfo->next) {
		if (!strcmp(ifinfo->ifname, ifname))
			break;
	}
	if (!ifinfo)
		return 1;

	/* get class info */
	if (clname) {
		for (class = ifinfo->cllist; class; class = class->next) {
			if (!strcmp(class->clname, clname))
				break;
		}
		if (!class && strcmp(clname, "null"))
			return 1;
	}
	memset((void *) &fltr_add, 0, sizeof(fltr_add));

	strncpy(fltr_add.cbq_iface.cbq_ifacename, ifname, IFNAMSIZ);
	fltr_add.cbq_iface.cbq_ifacelen = strlen(ifname);
	fltr_add.cbq_class_handle = class ? class->handle : NULL_CLASS_HANDLE;
	fltr_add.cbq_filter = *filt;

	if (cbqdebug > 5) {
		printf("%s: CBQ_ADD_FILTER, class.%x\n",
			ifname, fltr_add.cbq_class_handle);
		printf("Filter Destination Address: 0x%x\n",
			fltr_add.cbq_filter.daddr);
		printf("Filter Destination Port: 0x%x\n",
			fltr_add.cbq_filter.dport);
		printf("Filter Source Address: 0x%x\n",
			fltr_add.cbq_filter.saddr);
		printf("Filter Source Port: 0x%x\n",
			fltr_add.cbq_filter.sport);
		printf("Filter IP Protocol: 0x%x\n",
			fltr_add.cbq_filter.proto);
		printf("\n");
	}

	if (ifinfo->use_cbq && ioctl(cbqfd, CBQ_ADD_FILTER, &fltr_add) < 0) {
		perror("ioctl CBQ_ADD_FILTER");
		ifinfo->use_cbq = 0;
		return 1;
	}
	return 0;
}

static cbq_fltrinfo_t*
cbq_set_filter(cbq_flowinfo_t *flinfo, cbq_filt_t *cfilter)
{
	cbq_ifinfo_t   *ifinfo = flinfo->clinfo->ifinfo;
	struct cbq_add_filter fltr_add;
	cbq_fltrinfo_t *fltr_info;
	int i, j;

	memset((void *) &fltr_add, 0, sizeof(fltr_add));
	strncpy(fltr_add.cbq_iface.cbq_ifacename, ifinfo->ifname, IFNAMSIZ);
	fltr_add.cbq_iface.cbq_ifacelen = strlen(ifinfo->ifname);
	fltr_add.cbq_class_handle = flinfo->handle;
	fltr_add.cbq_filter = *cfilter;

	if (cbqdebug) {
		printf("%s: CBQ_ADD_FILTER, class.%x\n",
			ifinfo->ifname, fltr_add.cbq_class_handle);
		printf("Filter Destination Address: 0x%x\n",
			fltr_add.cbq_filter.daddr);
		printf("Filter Destination Port: 0x%x\n",
			fltr_add.cbq_filter.dport);
		printf("Filter Source Address: 0x%x\n",
			fltr_add.cbq_filter.saddr);
		printf("Filter Source Port: 0x%x\n",
			fltr_add.cbq_filter.sport);
		printf("Filter IP Protocol: 0x%x\n",
			fltr_add.cbq_filter.proto);
		printf("\n");
	}

	if (ifinfo->use_cbq && ioctl(cbqfd, CBQ_ADD_FILTER, &fltr_add) < 0) {
		perror("ioctl CBQ_ADD_FILTER");
		ifinfo->use_cbq = 0;
		return NULL;
	}
	if ((fltr_info = (cbq_fltrinfo_t *)
			malloc(sizeof(cbq_fltrinfo_t))) == NULL)
		return NULL;
	fltr_info->handle = fltr_add.cbq_filter_handle;
	fltr_info->next = flinfo->fltr_list;
	fltr_info->filter = *cfilter;
	fltr_info->flow = flinfo;
	flinfo->fltr_list = fltr_info;
	return fltr_info;
}

int
cbq_delfilter_all(cbq_flowinfo_t * flow_info)
{
	cbq_fltrinfo_t *fltr_info = flow_info->fltr_list;
	cbq_ifinfo_t   *ifinfo = flow_info->clinfo->ifinfo;
	cbq_fltrinfo_t *fltr_next;
	struct cbq_delete_filter fltr_delete;

	if (cbqdebug) {	
		fprintf(stderr, "cbq_delfilter_all() called\n");
	}

	while (fltr_info) {
		fltr_next = fltr_info->next;

		memset((void *) &fltr_delete, 0, sizeof(fltr_delete));
		strncpy(fltr_delete.cbq_iface.cbq_ifacename,
			ifinfo->ifname, IFNAMSIZ);
		fltr_delete.cbq_iface.cbq_ifacelen = strlen(ifinfo->ifname);
		fltr_delete.cbq_filter_handle = fltr_info->handle;
		if (ifinfo->use_cbq && ioctl(cbqfd, CBQ_DEL_FILTER,
				&fltr_delete) < 0) {
			perror("ioctl CBQ_DEL_FILTER");
			ifinfo->use_cbq = 0;
			return -1;
		}
		free(fltr_info);
		fltr_info = fltr_next;
	}
	flow_info->fltr_list = NULL;
	return 0;
}

int 
cbq_delfilter(cbq_fltrinfo_t *fltr)
{
	struct cbq_delete_filter fltr_delete;
	cbq_ifinfo_t   *ifinfo = fltr->flow->clinfo->ifinfo;
	
	memset((void *) &fltr_delete, 0, sizeof(fltr_delete));
	strncpy(fltr_delete.cbq_iface.cbq_ifacename,
		ifinfo->ifname, IFNAMSIZ);
	fltr_delete.cbq_iface.cbq_ifacelen = strlen(ifinfo->ifname);
	fltr_delete.cbq_filter_handle = fltr->handle;
	if (ifinfo->use_cbq && ioctl(cbqfd, CBQ_DEL_FILTER,
			&fltr_delete) < 0) {
		perror("ioctl CBQ_DEL_FILTER");
		ifinfo->use_cbq = 0;
		return -1;
	}
	free(fltr);
	return 0;
}


int
cbq_clear(char *ifname)
{
	struct cbq_interface 	cbqif;
	cbq_ifinfo_t  		*ifinfo;

	for (ifinfo = cbq_iflist; ifinfo; ifinfo = ifinfo->next) {
		if (!strcmp(ifinfo->ifname, ifname))
			break;
	}
	if (!ifinfo)
		return 1;
		
	/* clear the interface */
	memset((void *) &cbqif, 0, sizeof(cbqif));
	strncpy(cbqif.cbq_ifacename, ifname, IFNAMSIZ);
	cbqif.cbq_ifacelen = strlen(ifname);
	if (ifinfo->use_cbq && ioctl(cbqfd, CBQ_CLEAR_HIERARCHY, &cbqif) < 0) {
		perror("ioctl CBQ_CLEAR_HIERARCHY");
		ifinfo->use_cbq = 0;
		return 1;
	}
	return 0;
}

int
cbq_destroyall(void)
{
	cbq_ifinfo_t   *ifinfo = cbq_iflist, *ifnext;
	cbq_classinfo_t *clinfo, *clnext;

	if (cbqfd < 0) {
		fprintf(stderr, "cbq_destroyall: CBQ device is not open\n");
		return ;
	}

	for (clinfo = ifinfo->cllist; clinfo; clinfo = clnext) {
		clnext = clinfo->next;
		if (clinfo->clname)
			free(clinfo->clname);
		free(clinfo);
	}

	for (ifinfo = cbq_iflist; ifinfo; ifinfo = ifnext) {
		(void) cbq_disable(ifinfo->ifname);
		(void) cbq_clear(ifinfo->ifname);
		ifnext = ifinfo->next;
		if (ifinfo->ifname)
			free(ifinfo->ifname);
		free(ifinfo);
	}
}

cbq_ifinfo_t *
cbq_getifinfo(char *ifname)
{
	cbq_ifinfo_t   *ifinfo;

	/* get interface info */
	for (ifinfo = cbq_iflist; ifinfo; ifinfo = ifinfo->next) {
		if (!strcmp(ifinfo->ifname, ifname))
			break;
	}
	if (!ifinfo)
		return NULL;
	return (ifinfo);
}

cbq_acinfo_t *
cbq_getacinfo(char *ifname) 
{
	cbq_ifinfo_t   *ifinfo;

	/* get interface info */
	for (ifinfo = cbq_iflist; ifinfo; ifinfo = ifinfo->next) {
		if (!strcmp(ifinfo->ifname, ifname))
			break;
	}
	if (!ifinfo)
		return NULL;
	return &ifinfo->cd_class->acinfo;
}


int
cbq_getstats(char *ifname, class_stats_t *stats, int nclasses)
{
	struct cbq_getstats	gs;
	cbq_ifinfo_t		*ifinfo;

	for (ifinfo = cbq_iflist; ifinfo; ifinfo = ifinfo->next) {
		if (!strcmp(ifinfo->ifname, ifname))
			break;
	}

	if (!ifinfo)
		return (1);

	memset((void *) &gs, 0, sizeof (gs));
	strncpy(gs.iface.cbq_ifacename, ifname, IFNAMSIZ);
	gs.iface.cbq_ifacelen = strlen(ifname);
	gs.nclasses = nclasses;
	gs.stats = stats;

	if (ifinfo->use_cbq && ioctl(cbqfd, CBQ_GETSTATS, &gs) < 0) {
		perror("ioctl CBQ_GETSTATS");
		ifinfo->use_cbq = 0;
		return (1);
	}

	return (0);
}

void
cbq_report_stats(char *interface, class_stats_t *stats, float *pstats,
	int nclasses)
{
	int			i;
	cbq_acinfo_t		*ifinfo;
	int			bandwdith;
	static int		counter = 0;

	for (i = 0; i < nclasses; i++) {
		if (cbqdebug >= 10) {
			printf("%s.%x.%d: np%d, nb%d, o%d, b%d, d%d, oa %d, dlys %d, c %d, %f\n",
				interface, stats[i].handle, stats[i].depth,
				stats[i].npackets, stats[i].nbytes,
				stats[i].over, stats[i].borrows, stats[i].drops,
				stats[i].overactions, stats[i].delays,
				counter, pstats[i]);
		} else if (cbqdebug > 0) {
			printf("%s.%x.%d: %d, %f\n", interface,
				stats[i].handle, stats[i].depth, counter,
				pstats[i]);
		}

	}
	counter++;
}
